@@ -1,10 +1,10 @@
 ;; -*- no-byte-compile: t; lexical-binding: nil -*-
-(define-package "oai" "0.1"
-  "AI-LLM blocks for org-mode."
+(define-package "oai" "0.2"
+  "AI-LLM chat blocks for org-mode."
   '((emacs "29.1"))
   :url "https://github.com/Anoncheg1/emacs-oai"
-  :commit "268eda829b16b9000eebbcc30f021c47338a9757"
-  :revdesc "268eda829b16"
+  :commit "48831f36bf7262d33cec19fa6bcafad6b7edfb59"
+  :revdesc "48831f36bf72"
   :keywords '("org" "comm" "url" "link")
   :authors '((nil . "github.com/Anoncheg1,codeberg.org/Anoncheg"))
   :maintainers '((nil . "github.com/Anoncheg1,codeberg.org/Anoncheg")))
