@@ -73,8 +73,8 @@
 (require 'gv) ; for setf
 (require 'json) ; json-read, json-read-from-string, json-encode
 (require 'oai-debug)
-(require 'oai-block-tags)
 (require 'oai-block)
+(require 'oai-block-tags)
 (require 'oai-timers)
 (require 'oai-async1) ; for `oai-async1-plist-get'
 
@@ -141,7 +141,8 @@ If mode is not chat but completion, appropriate model should be set."
   :group 'oai)
 
 
-(defcustom oai-restapi-openai-known-chat-models '("gpt-4o-mini"
+(defcustom oai-restapi-openai-known-chat-models '("gpt-5-chat-latest"
+                                                  "gpt-4o-mini"
                                                   "gpt-4"
                                                   "gpt-4-32k"
                                                   "gpt-4-turbo"
@@ -160,6 +161,7 @@ If mode is not chat but completion, appropriate model should be set."
                                                   "o1-mini"
                                                   "o3"
                                                   "o3-mini"
+                                                  "o3-pro"
                                                   "o4-mini"
                                                   "chatgpt-4o-latest")
   "Alist of OpenAI chat models from https://platform.openai.com/docs/models."
@@ -251,17 +253,6 @@ that is why defined as lambda with marker.")
 (defvar-local oai-restapi--current-request-is-streamed nil
   "Whether we expect a streamed response or a single completion payload.")
 
-(defvar oai-restapi-after-chat-insertion-hook nil
-  "Hook that is called when a chat response is inserted.
-Note this is called for every stream response so it will typically only
-contain fragments.
-For STREAM executed for every word and one time with TYPE=\'end.
-For non-STREAM executed one time with TYPE=\'end.
-Arguments: type, role-text, pos, buffer
-- TYPE - simbol \='role, \='text or'end,
-- ROLE-TEXT - text or role name,
-- POS - position before text insertion
-- STREAM - stream mode or a single insertion.")
 
 ;; (defvar-local oai-restapi--currently-reasoning nil)
 
@@ -564,19 +555,28 @@ Useful for small max-tokens.
                 (< max-tokens 500))
            (format "Answer very short with in %d sentences or %d lines or less." (/ max-tokens 29) (* (/ max-tokens 29) 3) ))
           ((and (>= max-tokens 500)
-                (<= max-tokens 1000))
+                (< max-tokens 900))
            (format "Answer short in %d paragraphs or %d pages or less." (/ max-tokens 200) (ceiling (/ max-tokens 600.0)))))))
 
 ;; -=-= Prepare content
 
-(defun oai-restapi-prepare-content (element req-type sys-prompt sys-prompt-for-all-messages max-tokens)
+(defun oai-restapi-prepare-content (element &optional noweb-control req-type sys-prompt sys-prompt-for-all-messages max-tokens)
   "Get content of ai block of element in current buffer.
 Handle Tags, two types of REQ-TYPE and separation of PROMPT and MESSAGES.
 Return:
 - vector with messages for chat REQ-TYPE.
 - string for completion REQ-TYPE.
-Parameters ELEMENT REQ-TYPE SYS-PROMPT SYS-PROMPT-FOR-ALL-MESSAGES
-MAX-TOKENS described in `oai-restapi-request-prepare'."
+ELEMENT is result of `oai-block-p'.
+Optional arguments:
+
+If NOWEB-CONTROL is not-nil, activate expansion of noweb links in all
+user messages.
+If MAX-TOKENS is not-nil, and
+ `oai-restapi-add-max-tokens-recommendation' is set, recommendation
+ about MAX-TOKENS added to first system message.
+REQ-TYPE is completion, return string, otherwise prcess body for chat.
+Parameters REQ-TYPE SYS-PROMPT SYS-PROMPT-FOR-ALL-MESSAGES described in
+`oai-restapi-request-prepare', used only if non-nil."
   (oai--debug "oai-restapi-prepare-content %s" sys-prompt-for-all-messages)
   ;; (let ((content (oai-block-get-content element))) ; string - is block content
   (if (eql req-type 'completion)
@@ -586,17 +586,21 @@ MAX-TOKENS described in `oai-restapi-request-prepare'."
     (let* ((messages (oai-block--collect-chat-messages-at-point element
                                                                 sys-prompt
                                                                 sys-prompt-for-all-messages
-                                                                (when oai-restapi-add-max-tokens-recommendation
-                                                                  (oai-restapi--get-lenght-recommendation max-tokens))))
-           (messages (oai-restapi--modify-vector-content messages 'user #'oai-block-tags-replace))
-           (messages (oai-restapi--modify-vector-content messages 'user #'oai-block-tags--clear-properties))
+                                                                (when (and max-tokens oai-restapi-add-max-tokens-recommendation)
+                                                                      (oai-restapi--get-lenght-recommendation max-tokens))))
+           (messages (oai-restapi--modify-vector-content messages #'oai-block-tags-replace 'user))
+           (messages (oai-restapi--modify-vector-content messages #'oai-block-tags--clear-properties))
+           (messages (if noweb-control
+                         (oai-restapi--modify-vector-content messages
+                                                             (lambda (string) (org-babel-expand-noweb-references (list "markdown" string)))
+                                                             'user)
+                       messages)) ; noweb
            (messages (oai-block--pipeline oai-restapi-after-prepare-messages-hook messages)))
       messages))) ; return vector
 
 
 ;; -=-= Main
-
-(defun oai-restapi-request-prepare (req-type element sys-prompt sys-prompt-for-all-messages model max-tokens top-p temperature frequency-penalty presence-penalty service stream &optional info)
+(defun oai-restapi-request-prepare (req-type element noweb-control sys-prompt sys-prompt-for-all-messages model max-tokens top-p temperature frequency-penalty presence-penalty service stream &optional info)
   "Compose API request from data and start a server-sent event stream.
 Call `oai-restapi-request' function as a next step.
 Called from `oai-call-block' in main file.
@@ -604,6 +608,7 @@ ELEMENT org-element - is ai block, should be converted to market at
 once.
 REQ-TYPE symbol - is completion or chat mostly.  Set
   `oai-block--get-request-type'.
+noweb-control - if non-nill, noweb links should be expanded.
 SYS-PROMPT string - first system instruction as a string.
 SYS-PROMPT-FOR-ALL-MESSAGES from
   `oai-restapi-default-inject-sys-prompt-for-all-messages' variable.
@@ -619,9 +624,9 @@ PRESENCE-PENALTY integer - -2-2, lower less repeat concepts.
 SERVICE symbol or string - is the AI cloud service such as openai or
   azure-openai.
 STREAM string - as bool, indicates whether to stream the response."
-  (oai--debug "oai-restapi-request-prepare %s" sys-prompt-for-all-messages)
   (ignore info)
-  (let* ((content (oai-restapi-prepare-content element req-type sys-prompt sys-prompt-for-all-messages max-tokens))
+  (oai--debug "oai-restapi-request-prepare %s" sys-prompt-for-all-messages)
+  (let* ((content (oai-restapi-prepare-content element noweb-control req-type sys-prompt sys-prompt-for-all-messages max-tokens))
          (end-marker (oai-block--get-content-end-marker element))
          (callback (if (eql req-type 'completion) ; chat - ; set to oai-restapi--current-url-request-callback
                        ;; completion mode
@@ -635,8 +640,8 @@ STREAM string - as bool, indicates whether to stream the response."
                                                                                 t))
                        ;; else - not stream
                        (lambda (result) (oai-block--insert-single-response end-marker
-                                                                             (oai-restapi--get-single-response-text result)
-                                                                             t 'final))))))
+                                                                           (oai-restapi--get-single-response-text result)
+                                                                           t))))))
     ;; - Call and save buffer.
     (oai-timers--set
      (oai-restapi-request service model callback
@@ -667,9 +672,11 @@ Same to `oai-restapi--normalize-response' that used for stream.
 We use separate version, because streaming is complicated,
 but we will keep them interchangable.
 Result used for `oai-block--insert-single-response'.
+Error should be handled before calling this function in
+ `oai-restapi--maybe-show-openai-request-error'.
 Return text of message."
   (when response
-    ;; (oai--debug "oai-restapi--get-single-response-text response:" response)
+    (oai--debug "oai-restapi--get-single-response-text response:" response)
     (if-let ((err-obj (plist-get response 'error)))
         (let ((message (or (plist-get err-obj 'message)
                            err-obj)))
@@ -835,7 +842,7 @@ We count running ones in global integer only.
 
 For not stream url return event and hook `after-change-functions'
  triggered only after url buffer already kill, that is why we don't use
- this hook.
+ this hook. For not stream we process data directly in callback.
 Use argument SERVICE to find endpoint, MODEL as parameter to request."
   ;; - HTTP body preparation as a string
   (let ((url-request-extra-headers (oai-restapi--get-headers service))
@@ -870,27 +877,26 @@ Use argument SERVICE to find endpoint, MODEL as parameter to request."
             (lambda (events)
               ;; "Called within url-request-buffer after `after-change-functions'"
               (ignore events)
-              ;; called one time at error or at the end of all receiving.
-
-              (oai--debug "oai-restapi-request *url-retrieve callback*")
-
+              ;; debug
               (let (oai-restapi--url-buffer-last-position-marker)
                 (oai-restapi--debug-urllib (current-buffer)))
-
-              ;; (if (oai-restapi--maybe-show-openai-request-error) ; TODO: change to RESULT by global customizable option ; maybe use _events: (:error (error http 400)
-                  ;; for stream
-                  ;; (remove-hook 'after-change-functions #'oai-restapi--url-request-on-change-function t))
-                ;; Called for not stream, call `oai-restapi--current-url-request-callback'
-              (when (or nil ; oai-restapi--current-request-is-streamed
-                        (not (oai-restapi--maybe-show-openai-request-error))) ; t if error
-                (oai-restapi--url-request-on-change-function nil nil nil)) ; should be always called for stream
+              ;; error handling and not-stream insert
+              (unwind-protect
+                  (when (and (boundp 'url-http-end-of-headers) url-http-end-of-headers)
+                      (unless (oai-restapi--maybe-show-openai-request-error) ; t if error
+                        (unless stream
+                          (goto-char url-http-end-of-headers)
+                          ;; insert [ME]
+                          (funcall oai-restapi--current-url-request-callback
+                                   (oai-restapi--json-decode-not-streamed (buffer-substring-no-properties (point) (point-max))))
+                          ;; (funcall oai-restapi--current-url-request-callback nil)
+                          )))
 
 
-              ;; finally stop track buffer, error or not
-              ;; (oai--debug "Main request lambda" _events)
-              (oai-timers--interrupt-current-request (current-buffer) #'oai-restapi--stop-tracking-url-request)
-              ;; (oai-timers--interrupt-current-request (current-buffer) #'oai-restapi--interrupt-url-request)
-              ))))
+                ;; finally stop track buffer, error or not
+                (oai-timers--interrupt-current-request (current-buffer) #'oai-restapi--stop-tracking-url-request)
+                ;; (oai-timers--interrupt-current-request (current-buffer) #'oai-restapi--interrupt-url-request)
+                )))))
 
       (oai--debug "Main request after." url-request-buffer)
 
@@ -1051,8 +1057,8 @@ We store url-buf with marker of header in oai-timers.el"
   (with-current-buffer (marker-buffer header-marker)
     (let* (
           ;; prepare request - apply tags to message
-          (messages (oai-restapi--modify-vector-content messages 'user #'oai-block-tags-replace))
-          (messages (oai-restapi--modify-vector-content messages 'user #'oai-block-tags--clear-properties))
+          (messages (oai-restapi--modify-vector-content messages #'oai-block-tags-replace 'user))
+          (messages (oai-restapi--modify-vector-content messages #'oai-block-tags--clear-properties 'user))
           (messages (oai-block--pipeline oai-restapi-after-prepare-messages-hook messages)))
       (when (or (and retries (> retries 0))
                 (not retries))
@@ -1297,12 +1303,12 @@ Return nil if error."
 Arguments _BEG _END _LEN are not used. They are:
 the positions of the beginning and end of the range of changed text,
 and the length in chars of the pre-change text replaced by that range.
-
 Call `oai-restapi--current-url-request-callback' with data.
 After processing call `oai-restapi--current-url-request-callback' with nil.
 This  callback  here  is `oai-block--insert-stream-response'  for  chat  or
 `oai-block--insert-single-response' for completion.
-Called within `url-retrieve' buffer.
+Called within `url-retrieve' buffer, from `after-change-functions'
+ variable and from callback of `url-request-buffer'.
 Return JSOIN in plist format."
   (when (and (boundp 'url-http-end-of-headers)
              url-http-end-of-headers
@@ -1409,13 +1415,14 @@ Return JSOIN in plist format."
 
         ;; - Not-streamed
         ;; response is a single JSON object, no "data: " prefix. {"choices":[...
-        (when (not oai-restapi--current-request-is-streamed)
-          (oai--debug "oai-restapi--url-request-on-change-function NOT-STREAM: %s %s" (point) (point-max))
-          (let ((data (oai-restapi--json-decode-not-streamed (buffer-substring-no-properties (point) (point-max)))))
-            (when data
-              (funcall oai-restapi--current-url-request-callback data))
-            ;; - Done or Error
-            (funcall oai-restapi--current-url-request-callback nil)))))))
+        ;; (when (not oai-restapi--current-request-is-streamed)
+        ;;   (oai--debug "oai-restapi--url-request-on-change-function NOT-STREAM: %s %s" (point) (point-max))
+        ;;   (let ((data (oai-restapi--json-decode-not-streamed (buffer-substring-no-properties (point) (point-max)))))
+        ;;     (when data
+        ;;       (funcall oai-restapi--current-url-request-callback data))
+        ;;     ;; - Done or Error
+        ;;     (funcall oai-restapi--current-url-request-callback nil)))
+        ))))
 
 
 ;; -=-= Reporter & Requests interrupt functions
@@ -1569,10 +1576,11 @@ MAX-TOKEN-RECOMMENDATION SEPARATOR at `oai-block--collect-chat-messages'."
     idx))
 
 
-(defun oai-restapi--modify-vector-content (vec role new-content)
+(defun oai-restapi--modify-vector-content (vec new-content &optional role &rest rest)
   "Modify every string content of messages VEC that match role.
-ROLE is symbol.  NEW-CONTENT may be string or function that accept one
-string arguments and return new content."
+When ROLE is non-nil, it used to filter all such messages.
+NEW-CONTENT may be string or function that accept one
+string arguments and return new content, like `oai-block-tags-replace'."
   (unless (vectorp vec)
     (error "Not a vector"))
   (let ((i (1- (length vec)))
@@ -1584,19 +1592,25 @@ string arguments and return new content."
       (setq el (aref newvec i)) ; from 0 to len-1
       (setq content (plist-get el :content))
       ;; replace content for role
-      (when (eql role (plist-get el :role))
+      (when (or (and role
+                     (eql role (plist-get el :role)))
+                (not role))
         (setq content (if (functionp new-content)
-                          (funcall new-content content)
+                          (if rest
+                              (apply #'funcall new-content content rest)
+                            ;; else
+                            (funcall new-content content))
                         ;; else
                         new-content))
         (aset newvec i (plist-put (copy-sequence el) :content content)))
-      (setq i (1- i))) ; ++1
+      (setq i (1- i)))
     newvec))
 
 
-(defun oai-restapi--modify-last-user-content (vec new-content)
+(defun oai-restapi--modify-last-user-content (vec new-content &rest rest)
   "Replacing last \='user :content with NEW-CONTENT in VEC.
-NEW-CONTENT is either (string or function of old content).
+NEW-CONTENT is either (string or function of old content), like
+`oai-block-tags-replace'.
 Uses `oai-restapi--find-last-user-index`.
 Return new vector based on VEC.
 Used in `oai-restapi-request-prepare' to send history of conversation."
@@ -1608,8 +1622,10 @@ Used in `oai-restapi-request-prepare' to send history of conversation."
       (let* ((elt (aref newvec idx))
              (old-content (plist-get elt :content))
              (rep-content (if (functionp new-content)
-                              ;; oai-restapi--modify-last-user-content
-                              (funcall new-content old-content)
+                              (if rest
+                                  (apply #'funcall new-content old-content rest)
+                                ;; else
+                                (funcall new-content old-content))
                             new-content))
              (new-elt (plist-put (copy-sequence elt) :content rep-content)))
         (aset newvec idx new-elt)))
