@@ -1,11 +1,11 @@
-;;; oai.el --- AI-LLM blocks for org-mode -*- lexical-binding: t; -*-
+;;; oai.el --- AI-LLM chat blocks for org-mode -*- lexical-binding: t; -*-
 
 ;; Copyright (C) 2025 github.com/Anoncheg1,codeberg.org/Anoncheg
 ;; Author: <github.com/Anoncheg1,codeberg.org/Anoncheg>
 ;; Keywords: org, comm, url, link
 ;; URL: https://github.com/Anoncheg1/emacs-oai
-;; Package-Version: 0.1
-;; Package-Revision: 268eda829b16
+;; Package-Version: 0.2
+;; Package-Revision: 48831f36bf72
 ;; Created: 27 dec 2025
 ;; Package-Requires: ((emacs "29.1"))
 ;; Optional dependency: ((org-links "0.2"))
@@ -52,11 +52,21 @@
 ;; of features and setup instructions.
 ;;
 ;; Configuration:
-;; (add-to-list 'load-path "path/to/oai")
+;; (add-to-list 'load-path "path/to/oai") ; (optional)
 ;; (require 'oai)
 ;; (add-hook 'org-mode-hook #'oai-mode) ; oai.el
-;; (setq oai-restapi-con-token "xxx") ; oai-restapi.el
-;;
+;; (setq oai-restapi-con-token "xxx") ; oai-restapi.el (optional)
+
+;; Configuration, optional hooks:
+;; (add-hook 'oai-block-after-chat-insertion-hook
+;;   #'oai-optional-remove-distant-empty-lines-hook-function)
+;; (add-hook 'oai-block-after-chat-insertion-hook
+;;   #'oai-optional-remove-headers-hook-function)
+
+;; One hooks add space “ “ before lines that in ai block that looks
+;; like Org header, that fix confustion for Org logic.
+;; Another hook remove empty lines if there is too much of them in response.
+
 ;; You will need an OpenAI API key-token.
 ;; It can be stored in the format:
 ;;  "machine api.openai.com login oai password <your-api-key>"
@@ -68,15 +78,20 @@
 ;; - In block #+begin_ai..#+end_ai blocks:
 ;;     - C-c C-c - to send the text to the OpenAI API and insert a response
 ;;     - C-c . - to inspect raw data (and C-u C-c .)
-;;     - C-c C-. - to see url.el raw HTTP data (working only during request)
-;;     - M-h - mark ai block content
-;;     - C-u M-h - mark chat message
-;;     - C-c <backspace> - (kill-region-at-point) to remove the chat
-;;       part under point.  (oai-block.el)
-;;     - C-c m - set :max-tokens
+;;     - C-c C-.  - to see url.el raw HTTP data (working only during request)
+;;     - M-h - mark element in ai block (C-u M-h - mark chat message)
+;;     - C-c C-t - set :max-tokens
 ;; - in buffer with oai-mode enabled:
 ;;     - C-g - to stop all requsts.
 
+;; Customization:
+;; M-x customize-group RET oai
+;; M-x customize-group RET oai-faces
+
+;; Terms:
+;; - chat roles or prefixes - [AI]: [ME:]
+;; - parts or messages - major parts of chat with prefixes of roles
+
 
 ;; Other packages:
 ;; - Modern navigation in major modes https://github.com/Anoncheg1/firstly-search
@@ -108,17 +123,51 @@
 ;; - key to enable full Org highlighting? think about it
 ;; - fontify latex [[file:/usr/share/emacs/30.2/lisp/org/org.el::16097::(defun org-inside-latex-macro-p ()]]
 ;; [[file:/usr/share/emacs/30.2/lisp/textmodes/tex-mode.el::1277::(setq-local font-lock-defaults]]
+;; - small markdown mode on highlighting
+;; - simple Elisp function to ask LLM
+;; - add guide to use `oai-restapi-request' and with retries for simple ELisp LLM call and get result for TAB key and some place in buffer.
+;; - add option for tag to expand only the last user prompt or in all.
+;; - C-c C-k should jump to current begining of message, not next
+;; - add buttons: 1) generate button based on LLM answer 2) handle clicking.
+;; - default requst as one plist configuration
+;; - support for https://github.com/LionyxML/markdown-ts-mode
+;; - check big markdown-mode for insights for us.
+;; - stop previous request if new one called with all equal parameters
+;; - fill-paragraph should not break markdown quotes and bolds
+;; - add new line before ai answer
+;; - make font-lock better like in [[file:/usr/share/emacs/30.2/lisp/gnus/message.el
+;; ::1701::(defun message-font-lock-make-cited-text-matcher (level maxlevel)]]
+;; - make `oai-expand-block' executed with `org-babel-expand-src-block'.
+;; - provide place or hook to add custom expansion of link to one line for user defined mode
+;; - support vars as tags    https://orgmode.org/manual/Environment-of-a-Code-Block.html
+;; - write test for `oai-block-tags-get-content' and `oai-block-tags--get-content-at-point-org' and noweb parameter usage
+;; - add advanced forward section that check what type of region is
+;; active and do appropriate forward with preserving region
+;; - noweb evaluation with support of variables with some text. like <<call("as")>>
+;; - rebind keys to C-x C-a
+;; - function to replace "^[\s+]- **word1 [word2]:**" to "^^[\s+]- word1 [word2] :: " and highligh it.
+;; - fix highlight to highlight when there is only "#+end_ai"
+;; - create function that insert :max-token for given int
 ;;; Code:
 ;; Touch: Pain, water and warm.
 
 ;; -=-= includes
 (require 'oai-debug)
-(require 'oai-block-tags) ; `oai-block-tags-replace' for `oai-expand-block'
 (require 'oai-block)
+(require 'oai-block-tags) ; `oai-block-tags-replace' for `oai-expand-block'
 (require 'oai-restapi)
 (require 'oai-prompt) ; for `oai-prompt-request-chain'
 
-;; -=-= C-c C-c main interface
+;; -=-= Customs and groups
+(defgroup oai nil
+  "OAI package customization."
+  :group 'oai)
+
+(defgroup oai-faces nil
+  "Faces for OAI blocks."
+  :tag "OAI Faces"
+  :group 'oai)
+
 (defcustom oai-fontification-flag t
   "Non-nil means enable fontification for markdown and Org elements in block."
   :type 'boolean
@@ -136,6 +185,8 @@ parameters."
                 :tag "Property list (symbol => funcion)")
   :group 'oai)
 
+;; -=-= C-c C-c main interface
+
 (defun oai-ctrl-c-ctrl-c ()
   "Remove result and parse ai block header parameters.
 Returning t is Org requirement."
@@ -177,7 +228,9 @@ Return list of arguments args."
                                           oai-restapi-default-inject-sys-prompt-for-all-messages)) ; oai-restapi.el
          (sys-prompt (or (org-entry-get-with-inheritance "SYS") ; org
                          (oai-block--get-sys :info info ; oai-block.el
-                                             :default oai-restapi-default-chat-system-prompt)))) ; oai-restapi.el variable
+                                             :default oai-restapi-default-chat-system-prompt)))
+         (noweb-control (or (org-babel-noweb-p info :eval)
+                            (org-entry-get (point) "oai-noweb" t)))) ; oai-restapi.el variable
     ;; - Process Org params and call agent
     (oai-block--let-params info
                            ;; format: (variable optional-default type)
@@ -204,7 +257,7 @@ Return list of arguments args."
                                           ;; else
                                           model)))
                                  ;; return
-                                 (list element sys-prompt sys-prompt-for-all-messages ; message
+                                 (list element noweb-control sys-prompt sys-prompt-for-all-messages ; message
                                           model max-tokens top-p temperature frequency-penalty presence-penalty service stream ; model params
                                           info)))))
 
@@ -212,16 +265,18 @@ Return list of arguments args."
 (defun oai-expand-block-deep ()
   "Output almost RAW information about request with headers and messages.
 Return list of strings to print."
-  (seq-let (element sys-prompt sys-prompt-for-all-messages model max-tokens top-p temperature frequency-penalty presence-penalty service stream info) (oai-parse-org-header)
+  (seq-let (element noweb-control sys-prompt sys-prompt-for-all-messages model max-tokens top-p temperature frequency-penalty presence-penalty service stream info) (oai-parse-org-header)
     (let* ((req-type-completion (not (eq 'x (alist-get :completion info 'x))))
            ;; (req-type
            (messages (unless req-type-completion
                        ;; - split content to messages
-                       (oai-restapi--collect-chat-messages-at-point element
-                                                                    sys-prompt
-                                                                    sys-prompt-for-all-messages
-                                                                    (if oai-restapi-add-max-tokens-recommendation
-                                                                        (oai-restapi--get-lenght-recommendation max-tokens))))))
+                       (oai-restapi-prepare-content element noweb-control 'chat sys-prompt sys-prompt-for-all-messages max-tokens)
+                       ;; (oai-restapi--collect-chat-messages-at-point element
+                       ;;                                              sys-prompt
+                       ;;                                              sys-prompt-for-all-messages
+                       ;;                                              (if oai-restapi-add-max-tokens-recommendation
+                       ;;                                                  (oai-restapi--get-lenght-recommendation max-tokens)))
+                       )))
       (list
        (oai-restapi--get-endpoint messages service)
        (oai-restapi--get-headers service)
@@ -236,7 +291,6 @@ Return list of strings to print."
 			     :service service
 			     :stream stream)))))
 
-;;;###autoload
 (defun oai-expand-block (arg)
   "Show a temp buffer with what the ai block expands to.
 If there is ai block at current position in current buffer.
@@ -244,34 +298,45 @@ This is what will be sent to the api.  ELEMENT is the ai block.
 Like `org-babel-expand-src-block'.
 Set `help-window-select' variable to to t to get focus.
 When universal  ARG specifide  output more  raw information  splitted by
-messages."
+messages.
+Return expanded content if at current point of current buffer supported
+block was found, otherwise nil."
+  ; org-babel-expand-src-block put overlay with `org-src--make-source-overlay'
+  ; We add text properties in `oai-block-tags--replace-last-regex-smart'
   (interactive "P")
-  (when-let* ((element (oai-block-p)) ; oai-block.el
+  (when-let* ((element (oai-block-p)) ; (oai-block-tags--block-at-point))) ; oai-block.el
               (res-str (if arg
                            (pp-to-string (oai-expand-block-deep))
                          ;; else
-                         ;; (print (oai-block-tags-replace (oai-block-get-content element)))
-                         (let* (
-                                ;; get block
-                                (expanded (string-trim
-                                           (oai-block-get-content element))) ; oai-block.el
-                                ;; (expanded (oai-block-tags-replace expanded))
-                                ;; vector
-                                (expanded (oai-block--collect-chat-messages-from-string expanded))
-                                (expanded (oai-restapi--modify-vector-content expanded 'user #'oai-block-tags-replace))
-                                ;; one string
-                                (expanded (oai-block--stringify-chat-messages expanded)))
-                           expanded))))
+                         ;; (let* (
+                         ;;        ;; get block
+
+                         ;;        (expanded (string-trim
+                         ;;                   (oai-block-get-content element))) ; oai-block.el
+                         ;;        ;; (expanded (oai-block-tags-replace expanded))
+                         ;;        ;; vector
+                         ;;        (expanded (oai-block--collect-chat-messages-from-string expanded))
+                         ;;        (expanded (oai-restapi--modify-vector-content expanded #'oai-block-tags-replace 'user))
+                         ;;        ;; one string
+                         ;;        (expanded (oai-block--stringify-chat-messages expanded)))
+                         ;;   expanded))))
+                         (oai-block-tags-get-content element
+                                                     t ; tags-control
+                                                     t  ; noweb-control
+                                                     :eval ; noweb-context
+                                                      t ; links-only-last
+                                                      t ; not-clear-properties
+                                                     ))))
     (if (called-interactively-p 'any)
-        (let ((buf (get-buffer-create "*OAi Preview*")))
+        (let ((buf (get-buffer-create "*OAI Preview*")))
           (with-help-window buf (with-current-buffer buf
                                   (insert res-str)))
-          (switch-to-buffer buf))
+          (switch-to-buffer buf)
+          t)
       ;; else
       res-str)))
 
 ;; -=-= interactive fn: key C-g: keyboard quit
-
 (defun oai-keyboard-quit ()
   "Keyboard quit advice.
 - If there is an active region at current position in current buffer, do
@@ -296,9 +361,7 @@ messages."
               (call-interactively #'oai-restapi-stop-url-request) ; oai-restapi.el
             (error nil)))))))
 
-
 ;; -=-= interactive fn: M-x oai-toggle-debug
-;;;###autoload
 (defun oai-toggle-debug ()
   "Enable/disable debug."
   (interactive)
@@ -310,32 +373,83 @@ messages."
     (setq oai-debug-buffer   "*debug-oai*")
     (message "Enable oai debugging")))
 
-;; -=-= interactive aliases
-;;;###autoload
-(defalias 'oai-kill-region-at-point #'oai-block-kill-region-at-point)
-;;;###autoload
-(defalias 'oai-mark-at-point #'oai-block-mark-at-point)
-;;;###autoload
-(defalias 'oai-set-max-tokens #'oai-block-set-max-tokens)
-;;;###autoload
-(defalias 'oai-forward-section #'oai-block-forward-section)
+;; -=-= fn: Help function to rebind major mode with chaining
+(defun oai--call-next-remap-protected (command &optional seen)
+  "Call the next remapping of COMMAND, skipping any commands already in SEEN.
+If no further remappings found, calls COMMAND interactively if possible."
+  (let ((minor-mode-map-alist (cdr minor-mode-map-alist)))
+    (let ((binding (key-binding (vector 'remap command))))
+      (cond
+       ;; No binding found, or recursion, fallback to original
+       ((or (null binding) (memq binding seen))
+        (when (commandp command)
+          (call-interactively command)))
+       ;; Valid binding, try further
+       ((commandp binding)
+        (oai--call-next-remap-protected command (cons binding seen)))))))
 
-;; -=-= Fontify Markdown blocks and Tags - function for hook
+(defun oai--call-next-key-remap-protected (key &optional seen)
+  "Call the next binding of KEY, skipping handlers already in SEEN.
+If no further binding found, calls the major mode's or global binding.
+KEY is a string representing the keystroke.
+SEEN is a list of commands already called, used to prevent recursion."
 
-(defun oai-block--set-ai-keywords ()
-  "Hook, that Insert our fontify functions in Org font lock keywords."
-  ;; add fontify-ai-subblocks - markdown blocks and tables.
-  (when oai-fontification-flag
-    (setq org-font-lock-extra-keywords (oai-block--insert-after
-                                        org-font-lock-extra-keywords
-                                        (seq-position org-font-lock-extra-keywords '(org-fontify-meta-lines-and-blocks))
-                                        '(oai-block--font-lock-fontify-markdown-and-org)))
-    ;; add fontify-links
-    (setq org-font-lock-extra-keywords (oai-block--insert-after
-                                        org-font-lock-extra-keywords
-                                        (seq-position org-font-lock-extra-keywords '(org-fontify-meta-lines-and-blocks))
-                                        '(oai-block-tags--font-lock-fontify-links)))))
+  ;; Locally shadow minor-mode-map-alist to remove the highest-priority minor mode map.
+  (let ((minor-mode-map-alist (cdr minor-mode-map-alist)))
+    ;; Find the current binding for KEY after skipping the top minor mode.
+    (let ((binding (key-binding (kbd key) nil nil)))
+      (cond
+       ;; If no binding found or we've already seen this binding, try major mode and then global map.
+       ((or (null binding) (memq binding seen))
+        ;; Attempt to find the binding in the major mode's keymap.
+        (let* ((major-mode-map (current-local-map))
+               (binding-major (and major-mode-map (lookup-key major-mode-map (kbd key)))))
+          (if (commandp binding-major)
+              ;; If found and it's a command, call interactively.
+              (call-interactively binding-major)
+            ;; Otherwise, try the global map for the key.
+            (let ((global-binding (key-binding (kbd key) t t)))
+              (if (commandp global-binding)
+                  (call-interactively global-binding)
+                ;; If no valid binding anywhere, notify the user.
+                (message "No valid binding for %s" key))))))
+       ;; If binding is a command, recursively try to find the next remapped binding,
+       ;; and add this binding to SEEN for recursion protection.
+       ((commandp binding)
+        (oai--call-next-key-remap-protected key (cons binding seen)))
+       ;; Handle the case where binding is not a command (function, lambda, etc.).
+       (t
+        (message "Binding for %s is not a command" key))))))
 
+;; -=-= interactive fns: Org keys remapings
+(defun oai-mark-at-point-org (&optional arg)
+  "Mark element of ai block, large we every next execution.
+If optional argument ARG is non-nil, mark current message of chat."
+  (interactive "P")
+  (if (oai-block-p)
+      (if arg
+          (progn (deactivate-mark)
+                 (oai-block-mark-chat-message-at-point)
+                 (message "Chat message"))
+        ;; else
+        (call-interactively #'oai-block-mark-at-point))
+    ;; else
+    (oai--call-next-remap-protected #'org-mark-element)))
+
+(defun oai-expand-block-org ()
+  "Show a temp buffer with what the ai block expands to."
+  (interactive)
+  (if (not (call-interactively #'oai-expand-block))
+    ;; else
+    (oai--call-next-key-remap-protected "C-c .")))
+
+(defun oai-set-max-tokens-org ()
+  "Jump to header of ai block and set max-tokens."
+  (interactive)
+  (if (oai-block-p)
+      (oai-block-set-block-parameter ":max-tokens" oai-restapi-default-max-tokens)
+    ;; else
+    (oai--call-next-key-remap-protected "C-c C-t")))
 
 ;; -=-= interactive fn: Summarize *TODO:*
 ;; (defun oai-summarize ()
@@ -355,18 +469,44 @@ messages."
 ;;           (while
 
 
+;; -=-= Minor mode: keymap
+;;;###autoload
+(defvar-keymap oai-mode-map
+  :repeat nil
+  :parent nil
+  "<remap> <org-next-visible-heading>" #'oai-block-next-message ; C-c C-n todo make org
+  "<remap> <org-previous-visible-heading>" #'oai-block-previous-message ; C-c C-p todo make org
+  "<remap> <org-mark-element>" #'oai-mark-at-point-org ; M-h
+  "C-c ." #'oai-expand-block-org
+  "C-c C-." #'oai-open-request-buffer
+  "C-c C-t" #'oai-set-max-tokens-org)
+
+;; -=-= Minor mode: hook - Fontify Markdown blocks and Tags - function for hook
+
+(defun oai-block--set-ai-keywords ()
+  "Hook, that Insert our fontify functions in Org font lock keywords."
+  ;; add fontify-ai-subblocks - markdown blocks and tables.
+  ;; Put in order to `org-font-lock-keywords': (oai-block--font-lock-fontify-markdown-and-org) (oai-block-tags--font-lock-fontify-links) (oai-block--font-lock-fontify-markdown-blocks)
+  (when oai-fontification-flag
+    ;; 3) fontify markdown blocks (and clear small)
+    (setq org-font-lock-extra-keywords (oai-block--insert-after
+                                        org-font-lock-extra-keywords
+                                        (seq-position org-font-lock-extra-keywords '(org-fontify-meta-lines-and-blocks))
+                                        '(oai-block--font-lock-fontify-markdown-blocks)))
+    ;; 2) fontify-links (and clear small)
+    (setq org-font-lock-extra-keywords (oai-block--insert-after
+                                        org-font-lock-extra-keywords
+                                        (seq-position org-font-lock-extra-keywords '(org-fontify-meta-lines-and-blocks))
+                                        '(oai-block-tags--font-lock-fontify-links)))
+    ;; 1) fontify small elements
+    (setq org-font-lock-extra-keywords (oai-block--insert-after
+                                        org-font-lock-extra-keywords
+                                        (seq-position org-font-lock-extra-keywords '(org-fontify-meta-lines-and-blocks))
+                                        '(oai-block--font-lock-fontify-markdown-and-org)))))
+
 ;; -=-= Minor mode
-(defvar oai-mode-map (make-sparse-keymap)
-  "Keymap for `oai-mode'.")
-
-(let ((map oai-mode-map))
-  (define-key map (kbd (string-join (list "C-c" " <backspace>"))) #'oai-kill-region-at-point) ; oai-block.el
-  (define-key map (kbd "M-h") #'oai-mark-at-point) ; oai-block.el
-  (define-key map (kbd "C-c C-.") #'oai-open-request-buffer) ; oai-restapi.el
-  (define-key map (kbd "C-c .") #'oai-expand-block) ; oai-block.el
-  (define-key map (kbd (concat "C-c " "m")) #'oai-set-max-tokens)) ; oai-block.el
-
 
+;;;###autoload
 (define-minor-mode oai-mode
   "Minor mode for `org-mode' integration with the OpenAI API."
   :init-value nil
@@ -389,7 +529,7 @@ messages."
     (org-set-font-lock-defaults)
     (font-lock-refresh-defaults)))
 
-;;;###autoload
+
 (defun oai-open-request-buffer ()
   "Opens the url request buffer for ai block at current position."
   (interactive)
@@ -413,7 +553,7 @@ messages."
         ;; else
         (message "No url buffer found"))
   ;; - else - no element - call original Org key
-  (call-interactively (lookup-key org-mode-map (kbd "C-c ?")))))
+  (oai--call-next-key-remap-protected "C-c C-.")))
 
 ;; -=-= Minor mode - string line
 (defvar oai-mode-line-string "")
